#include "config.h"                 // clock and protothreads configure
#include "pt_cornell_1_2_2.h"       // threading library

#include "tft_master.h"             // TFT graphics library
#include "tft_gfx.h"                // TFT graphics library
#include <stdlib.h>                 // need for rand function
#include <math.h>                   // for math operations
#include "GPS_UART.h"               // for UART setup
#include "i2cIMU.h"

#define m_thresh 5

// string buffers
char buffer[60]; // TFT buffer
char buffer32[32]; // SPI2 buffer
int motion = 0;

//==========================IMU=================================================
static float xbuf[20];
static float ybuf[20];
static float zbuf[20];
static int timebuf [20];

static float v_x;
static float v_y;
static float v_z;
static float vel;

signed short acc_x, acc_y, acc_z;
volatile signed short p_acc_x, p_acc_y, p_acc_z;

#define EnablePullDownA(bits) CNPUACLR=bits; CNPDASET=bits; // TODO: need this?

// ========= SPI Definitions ===================================================
volatile SpiChannel spiChn = SPI_CHANNEL2; // the SPI channel to use
volatile int spiClkDiv = 8; // 20 MHz max speed for this DAC

// ========= Thread Structures =================================================
// note that UART input and output are threads
static struct pt pt_gps, pt_imu, pt_wifi, pt_input, pt_output;

// ========= System Timer Var ==================================================
int sys_time_seconds;

volatile int accelbuf;
volatile int accelData;
volatile int junk;

// ========= GPS UART Thread ===================================================
// takes in an NMEA sentence from the GPS module over UART

static PT_THREAD(protothread_gps(struct pt *pt)) {
    PT_BEGIN(pt);
    while (1) {

        // spawn UART input thread until valid NMEA sentence is received
        PT_SPAWN(pt, &pt_input, PT_GetSerialBufferGPS(&pt_input));

        // if received sentence is in GPGGA format
        if (GPGGA == 1) {
            GPGGA = 0; //reset GPGGA flag
            parse(PT_term_buffer_GPS); //parse the GPGGA sentence
        } //end if

        //get the location and the nearby landmark
        gps_get_region();

        PT_YIELD_TIME_msec(100);
    } //end while(1)
    PT_END(pt);
} //end of GPS UART Thread

//============ HRM =============================================================
// Volatile Variables, used in the interrupt service routine!
volatile int BPM = 0; // int that holds raw Analog in 0. updated every 2mS
volatile int pulseSignal; // holds the incoming raw data
volatile int IBI = 600; // int that holds the time interval between beats! Must be seeded!
volatile char Pulse = 0; // "True" when User's live heartbeat is detected. "False" when not a "live beat".
volatile char QS = 0; // becomes true when Arduoino finds a beat.

volatile int rate[10]; // array to hold last ten IBI values
volatile unsigned long sampleCounter = 0; // used to determine pulse timing
volatile unsigned long lastBeatTime = 0; // used to find IBI
volatile int P = 512; // used to find peak in pulse wave, seeded
volatile int T = 512; // used to find trough in pulse wave, seeded
volatile int thresh = 530; // used to find instant moment of heart beat, seeded
volatile int amp = 0; // used to hold amplitude of pulse waveform, seeded
volatile char firstBeat = 1; // used to seed rate array so we startup with reasonable BPM
volatile char secondBeat = 0; // used to seed rate array so we startup with reasonable BPM

volatile int regVal;
volatile int ISR_end;


// ================== HRM ISR ==================================================

void __ISR(_TIMER_2_VECTOR, ipl2) Timer2Handler(void) {
    mT2ClearIntFlag();

    pulseSignal = ReadADC10(0);
    AcquireADC10(); // not needed if ADC_AUTO_SAMPLING_ON below

    sampleCounter += 2;

    int N;
    N = sampleCounter - lastBeatTime;

    //  find the peak and trough of the pulse wave
    if (pulseSignal < thresh && N > (IBI / 5)*3) { // avoid dichrotic noise by waiting 3/5 of last IBI
        if (pulseSignal < T) { // T is the trough
            T = pulseSignal; // keep track of lowest point in pulse wave
        }
    }

    if (pulseSignal > thresh && pulseSignal > P) { // thresh condition helps avoid noise
        P = pulseSignal; // P is the peak
    } // keep track of highest point in pulse wave

    // LOOK FOR THE HEART BEAT
    // signal surges up in value every time there is a pulse
    if (N > 250) { // avoid high frequency noise
        if ((pulseSignal > thresh) && (Pulse == 0) && (N > (IBI / 5)*3)) {
            Pulse = 1; // set the Pulse flag when we think there is a pulse
            //digitalWrite(blinkPin,HIGH);                // turn on pin 13 LED
            IBI = sampleCounter - lastBeatTime; // measure time between beats in mS
            lastBeatTime = sampleCounter; // keep track of time for next pulse

            int i;
            if (secondBeat) { // if this is the second beat, if secondBeat == TRUE
                secondBeat = 0; // clear secondBeat flag
                for (i = 0; i <= 9; i++) { // seed the running total to get a realisitic BPM at startup
                    rate[i] = IBI;
                }
            }

            if (firstBeat) { // if it's the first time we found a beat, if firstBeat == TRUE
                firstBeat = 0; // clear firstBeat flag
                secondBeat = 1; // set the second beat flag
                //sei();                               // enable interrupts again
                return; // IBI value is unreliable so discard it
            }

            // keep a running total of the last 10 IBI values
            int runningTotal = 0; // clear the runningTotal variable
            //
            for (i = 0; i <= 8; i++) { // shift data in the rate array
                rate[i] = rate[i + 1]; // and drop the oldest IBI value
                runningTotal += rate[i]; // add up the 9 oldest IBI values
            }

            rate[9] = IBI; // add the latest IBI to the rate array
            runningTotal += rate[9]; // add the latest IBI to runningTotal
            runningTotal = runningTotal / 10; // average the last 10 IBI values
            BPM = 60000 / runningTotal; // how many beats can fit into a minute? that's BPM!
            QS = 1; // set Quantified Self flag
            // QS FLAG IS NOT CLEARED INSIDE THIS ISR
        }
    }

    if (pulseSignal < thresh && Pulse == 1) { // when the values are going down, the beat is over
        //digitalWrite(blinkPin,LOW);            // turn off pin 13 LED
        Pulse = 0; // reset the Pulse flag so we can do it again
        amp = P - T; // get amplitude of the pulse wave
        thresh = amp / 2 + T; // set thresh at 50% of the amplitude
        P = thresh; // reset these for next time
        T = thresh;
    }

    if (N > 2500) { // if 2.5 seconds go by without a beat
        thresh = 530; // set thresh default
        P = 512; // set P default
        T = 512; // set T default
        lastBeatTime = sampleCounter; // bring the lastBeatTime up to date
        firstBeat = 1; // set these to avoid noise
        secondBeat = 0; // when we get the heartbeat back
    }
}

// ================== IMU Thread ===============================================
// parse IMU data to calculate speed

static unsigned int time_diff;

static PT_THREAD(protothread_imu(struct pt *pt)) {
    PT_BEGIN(pt);
    while (1) {
        acc_x = i2c1ConsecutiveRead(ACCEL_DATA_X_LSB_ADDR, 2);
        acc_y = i2c1ConsecutiveRead(ACCEL_DATA_Y_LSB_ADDR, 2);
        acc_z = i2c1ConsecutiveRead(ACCEL_DATA_Z_LSB_ADDR, 2);

        if (abs((int) p_acc_x - (int) acc_x) > m_thresh |
                abs((int) p_acc_y - (int) acc_y) > m_thresh |
                abs((int) p_acc_z - (int) acc_z) > m_thresh) motion = 1;
        else motion = 0;

        p_acc_x = acc_x;
        p_acc_y = acc_y;
        p_acc_z = acc_z;

        int i;
        for (i = 0; i < 19; i++) {
            xbuf[i + 1] = xbuf[i];
            ybuf[i + 1] = ybuf[i];
            zbuf[i + 1] = zbuf[i];
            timebuf[i + 1] = timebuf[i];
        }

        timebuf[0] = PT_GET_TIME();

        xbuf[0] = acc_x;
        ybuf[0] = acc_y;
        zbuf[0] = acc_z;
        time_diff = timebuf[0] - timebuf[19];
        if (time_diff != 0) {
            v_x = (xbuf[0] - xbuf[19]) / time_diff; // - xbuf[0]); // 2msec, 20 samples
            v_y = (ybuf[0] - ybuf[19]) / time_diff; // 2msec, 20 samples
            v_z = (zbuf[0] - zbuf[19]) / time_diff; // 2msec, 20 samples
            vel = sqrt(v_x * v_x + v_y * v_y + v_z * v_z);
        }
        PT_YIELD_TIME_msec(100);

    } //end while(1)
    PT_END(pt);
} //end of IMU thread

// ========= WiFi Thread =======================================================
// sends stats to the Huzzah over UART1

static PT_THREAD(protothread_wifi(struct pt *pt)) {
    PT_BEGIN(pt);
    while (1) {

        sprintf(PT_send_buffer, "TIME:\t%2d:%2d:%2d\n", GPS_time_h, GPS_time_m, GPS_time_s);
        PT_SPAWN(pt, &pt_output, PutSerialBuffer(&pt_output));

        sprintf(PT_send_buffer, "LAT:\t%9f\n", GPS_curLat);
        PT_SPAWN(pt, &pt_output, PutSerialBuffer(&pt_output));

        sprintf(PT_send_buffer, "LONG:\t%9f\n", GPS_curLon);
        PT_SPAWN(pt, &pt_output, PutSerialBuffer(&pt_output));

        sprintf(PT_send_buffer, "VEL:\t%3f\n", (float) vel);
        PT_SPAWN(pt, &pt_output, PutSerialBuffer(&pt_output));

        sprintf(PT_send_buffer, "BPM:\t%3d\n", BPM);
        PT_SPAWN(pt, &pt_output, PutSerialBuffer(&pt_output));

        sprintf(PT_send_buffer, "\n\0");
        PT_SPAWN(pt, &pt_output, PutSerialBuffer(&pt_output));

        PT_YIELD_TIME_msec(2000);

    } //end while(1)
    PT_END(pt);
} //end of WiFi thread

// ========= Main ==============================================================

void main(void) {

    ANSELA = 0;
    ANSELB = 0; //TODO: need this?

    /* Timer 2 Setup */
    //turn on Timer2; internal clock source, prescalar 8, toggle rate 500
    //at 30 MHz peripheral bus clock, 60 counts is 2 microseconds
    //2000 is 20 ksamples/sec
    OpenTimer2(T2_ON | T2_SOURCE_INT | T2_PS_1_8, 10000);

    //turn on Timer2 interrupt; interrupt priority = 2, clear interrupt flag
    ConfigIntTimer2(T2_INT_ON | T2_INT_PRIOR_2);
    mT2ClearIntFlag();

    // ADC setup
    CloseADC10(); // ensure the ADC is off before setting the configuration
    //  configure to sample AN5 and AN1 on MUX A and B AN1 = RA1, AN5 = RB3
    SetChanADC10(ADC_CH0_NEG_SAMPLEA_NVREF | ADC_CH0_POS_SAMPLEA_AN0);
    OpenADC10(PARAM1, PARAM2, PARAM3, PARAM4, PARAM5); // configure ADC using the parameters defined above
    EnableADC10(); // Enable the ADC

    // SDA = Pin 18
    // SCL = Pin 17
    // I2C1CONbits.SCLREL = 1;
    // I2C1CONbits.DISSLW = 1; // Slew Rate disabled for 100KHz mode
    // I2C1CONbits.ON = 1;     // Enable I2C1 module
    // I2C1BRG = 93;
    // calculated as follows: BRG = (Fpb / 2 / baudrate) - 2.
    // OPERATION_MODE_IMUPLUS chooses the sensor to operate as an IMU
    OpenI2C1(I2C_ON | I2C_SLW_DIS, 93);
    configIMU(OPERATION_MODE_IMUPLUS);

    /* Configure Protothreads */
    //turns off UART support and debugger pin, unless defines set
    PT_setup();

    //setup system wide interrupts
    INTEnableSystemMultiVectoredInt();

    //init protothreads
    PT_INIT(&pt_gps);
    PT_INIT(&pt_imu);
    PT_INIT(&pt_wifi);

    /* Schedule Protothreads */
    while (1) {
        PT_SCHEDULE(protothread_gps(&pt_gps));
        PT_SCHEDULE(protothread_imu(&pt_imu));
        PT_SCHEDULE(protothread_wifi(&pt_wifi));
    }
} //end main

// ========= IMU Helper Functions ==============================================

/*****************************************************************************
 * Initialization of the IMU
 *
 * This function brings the IMU into "mode" operation mode to allow for access
 * of predefined register values. Registers are:
 *****************************************************************************/
void configIMU(unsigned char mode) {
    // start the transmission sequence
    StartI2C1();
    IdleI2C1();

    // write the device address with R/W as 0
    MasterWriteI2C1(DEVICE_ADDRESS << 1);
    IdleI2C1();
    // make sure slave acknowledges the write
    if (I2C1STATbits.ACKSTAT) return;

    // write the address of the operation mode register
    MasterWriteI2C1(OPR_MODE_ADDR);
    IdleI2C1();
    // make sure slave acknowledges the write
    if (I2C1STATbits.ACKSTAT) return;

    // we pick the IMU mode to get the fusion data
    // this allows us to get the degrees relative to the IMU for pitch, yaw,
    // and roll
    // this writes mode to the operation register
    MasterWriteI2C1(mode); // pick mode
    IdleI2C1();
    // make sure slave acknowledges the write
    if (I2C1STATbits.ACKSTAT) return;

    // stop the transmission sequence
    StopI2C1();
    IdleI2C1();
}

/*****************************************************************************
 * Read from I2C
 *
 * This allows the PIC to read values from the IMU registers.
 * Arguments:
 * addr - this is the register address you want to read from in the IMU
 *****************************************************************************/
signed char i2c1Read(unsigned char addr) {
    unsigned char data;

    // start the transmission sequence
    StartI2C1();
    IdleI2C1();
    // write the device address with R/W as 0
    MasterWriteI2C1(DEVICE_ADDRESS << 1);
    IdleI2C1();
    // make sure slave acknowledges the write
    if (I2C1STATbits.ACKSTAT) return;

    // write the address of the register you want to read from
    MasterWriteI2C1(addr);
    IdleI2C1();
    // make sure slave acknowledges the write
    if (I2C1STATbits.ACKSTAT) return;

    //initiate repeated start
    StartI2C1();
    IdleI2C1();
    // rewrite the device address with R/W as 1 this time
    MasterWriteI2C1((DEVICE_ADDRESS << 1) | 1); // Issue Read Command
    IdleI2C1();
    // make sure slave acknowledges the write
    if (I2C1STATbits.ACKSTAT) return;

    // read the data from the specified register and nack to end transmission
    data = MasterReadI2C1();
    NotAckI2C1();
    // stop the transmission sequence
    StopI2C1();
    IdleI2C1();

    return data;
}

/*****************************************************************************
 * Consecutive Read from I2C
 *
 * This allows the PIC to read from consecutive registers from the IMU.
 * Arguments:
 * startReg - this is the first register you want to read from
 * num - this is the number of consecutive registers you want to read from
 *****************************************************************************/
signed short i2c1ConsecutiveRead(unsigned char startReg, unsigned char num) {
    // the first and second bytes read
    unsigned char lsb;
    signed char msb;
    // start the transmission sequence
    StartI2C1();
    IdleI2C1();
    // write the device address with R/W as 0
    MasterWriteI2C1(DEVICE_ADDRESS << 1);
    IdleI2C1();
    // make sure slave acknowledges the write
    if (I2C1STATbits.ACKSTAT) return;

    // write the first register you want to read from
    MasterWriteI2C1(startReg);
    IdleI2C1();
    // make sure slave acknowledges the write
    if (I2C1STATbits.ACKSTAT) return;

    //initiate repeated start
    StartI2C1();
    IdleI2C1();
    // write the device address again but this time with R/W as 1
    MasterWriteI2C1((DEVICE_ADDRESS << 1) | 1);
    IdleI2C1();
    // make sure slave acknowledges the write
    if (I2C1STATbits.ACKSTAT) return;

    // read the data for the first byte and send ack to slave to read consecutively
    lsb = MasterReadI2C1();
    AckI2C1();
    IdleI2C1();
    // read the data for the second byte and send nack to end transmission
    msb = MasterReadI2C1();
    NotAckI2C1();
    // stop the transmission sequence
    StopI2C1();
    IdleI2C1();
    // return the two bytes in the form of a signed short
    return (msb << 8) | lsb;
}


// ========= End ===============================================================